/*
 *	XENIX V CTCLIB.C module
 *
 *	Due to a bug in the non-exclusive record locks in System V Release 2.0
 *	(RLCK & NBRLCK which are described as the same as exclusive locks,
 *	but reads are permitted; and in fact allow more than one process
 *	to own the lock and allow an owner to update the record when there
 *	are no other owners - hence they are neither write locks nor read
 *	locks), we have provided two locking modes:
 *
 *		1) REGULAR - for use when advisory/exclusive locks are
 *			     available; and
 *		2) PHANTOM - otherwise.
 *
 *	c-tree does not require exclusive locks (which deny both read and
 *	write permission to other processes), but does want write locks: only
 *	one process can lock the record, but other processes may read the
 *	record.
 *
 *	Use REGULAR (by commenting out the #define PHANTOM) when exclusive
 *	are available in the adivory mode (i.e., a process can read the locked
 *	region if it does not ask for a lock which is how c-tree is designed.)
 *
 *	In the PHANTOM mode, exclusive locks are place on a dummy file so 
 *	that all processes can still read locked regions of the real files.
 *	In particular, the zeroth data file of each ISAM parameter
 *	file should be a dummy file, have a record length of 128 bytes, 
 * 	and file mode parameters indicating a SHARED, PERMANENT, FIXED length 
 *	file. Then, instead of actually locking the files in use, c-tree will
 *	place locks on corresponding regions of the dummy file. 
 *
 *	This program is the CONFIDENTIAL and PROPRIETARY property 
 *	of FairCom(R) Corporation. Any unauthorized use, reproduction or
 *	transfer of this program is strictly prohibited.
 *
 *      Copyright (c) 1985, 1986, 1987, 1988, 1989 FairCom
 *	(Subject to limited distribution and
 *	 restricted disclosure only.)
 *	*** ALL RIGHTS RESERVED ***
 *
 *	4006 West Broadway
 *	Columbia, MO 65203
 *
 *
 *	c-tree(R)	Version 4.3
 *			Release C
 *			February 7, 1989 17:30
 *
 */

#include "ctstdr.h"		/* standard i/o header 		*/
#undef   EXTERN
#define  EXTERN /* */
#include "ctoptn.h"		/* c-tree configuration options */
#include "cterrc.h"		/* c-tree error codes		*/
#include "ctstrc.h"		/* c-tree data structures	*/
#include "ctgvar.h"		/* c-tree global variables	*/
#include <sys/types.h>
#include <sys/locking.h>

#define PHANTOM	   /* delete when exclusive/advisory locks are available */

#ifdef  PHANTOM
#define LEFT_SHIFT	24	/* controls dummy lock capacity */
				/* 25 => 62  files: 32M fixed records or */
				/*		    32MB var len file    */
				/* 24 => 126 files: 16M fixed records or */
				/*		    16MB var len file    */
				/* 23 => 254 files:  8M fixed records or */
				/*		     8MB var len file    */
				/* 22 => 510 files:  4M fixed records or */
				/*		     4MB var len file    */
#endif

COUNT ctseek(),uerr();
   
RNDFILE mbopen(ctnum,opmode)
PFAST CTFILE *ctnum;
COUNT opmode;
{
	RNDFILE retval;
	int	acflag;
#ifndef CTSERVER
	int	i;
	long lseek();
#endif

	COUNT vtclose();

	ctnum->sekpos = 0L;
	acflag	      = BUPDATE;
	if (opmode & READFIL)
		acflag = O_RDONLY;

	if (!(opmode & PERMANENT) && ct_numvfil >= MAXVFIL)
		vtclose();

	if ((retval = open(ctnum->flname,acflag)) < 0)
		if (vtclose() == YES)
			retval = open(ctnum->flname,acflag);

#ifdef CTSERVER
	if (retval > -1 && locking(retval,LK_NBLCK,0L)) {
		close(retval);
		return(-1);
	}
#else
	/* obtain lock on entire file if exclusive open */
	if (retval > -1 && !(opmode & NONEXCLUSIVE)) {
		lseek(retval,ctnum->sekpos = HDRSIZ,0);
		if (locking(retval,LK_NBLCK,0L)) {
			close(retval);
			return(-1);
		}
	} else if (retval > -1 && (opmode & NONEXCLUSIVE)) {
		/* check file type to see if index: no lock on index */
		if (read(retval,ctnum,HDRSIZ) != HDRSIZ) {
			close(retval);
			return(-1);
		} else {
#ifdef UNIFRMAT
			revhdr(ctnum);
#endif
			ctnum->sekpos = HDRSIZ;
		}
		if (ctnum->clstyp != IDX_CLOSE) {
			for (i = HDRSIZ; i < 128; i++) {
				lseek(retval,ctnum->sekpos = i,0);
				if (locking(retval,LK_NBLCK,1L) == 0)
					break;
			}
			if (i >= 128) {
				close(retval);
				return(-1);
			}
		}
	}
#endif

	if (!(opmode & PERMANENT) && retval >= 0)
		ct_numvfil++;

	return(retval);
}

/* ------------------------------------------------------------ */

RNDFILE mbcrat(ctnum)
PFAST CTFILE *ctnum;
{
	RNDFILE retval;

	COUNT vtclose();

	ctnum->sekpos = 0L;
	if (!(ctnum->flmode & PERMANENT) && ct_numvfil >= MAXVFIL)
		vtclose();

	if ((retval = creat(ctnum->flname,BCREATE)) < 0)
		if (vtclose() == YES)
			retval = creat(ctnum->flname,BCREATE);

	if (!(ctnum->flmode & PERMANENT) && retval >= 0)
		ct_numvfil++;

	return(retval);
}

/* ------------------------------------------------------------ */

COUNT mbsave(ctnum)
PFAST CTFILE *ctnum;
{
	sync();
	return(NO_ERROR);
}

VOID flushdos(datno)
COUNT	      datno;
{
	sync();
}

#ifndef CTSERVER

/* --------------------------------------------------------------------
   LOCK index node
 */

COUNT LOCK(node,knum)	/* node == 0  => header */
LONG node;
PFAST KEYFILE *knum;
{

/*
 * c-tree node locking protocol is guaranteed to be deadlock free. Therefore,
 * the lock request allows the process to "sleep" if the lock is not
 * immediately available.
 */

#ifdef FPUTFGET
#ifndef PHANTOM
	knum -= knum->kmem;
	if (!(knum->flmode & NONEXCLUSIVE))
		return(NO_ERROR);

	if (ctseek(knum,node))
		return(uerr(SEEK_ERR));
	if (locking(knum->fd,LK_LOCK,1L))
		return(uerr(LNOD_ERR));
	else {
		knum->lokcnt++;
		return(NO_ERROR);
	}
#else
	knum -= knum->kmem;
	if (!(knum->flmode & NONEXCLUSIVE))
		return(NO_ERROR);

/*
 * Notice how a lock region is converted to a phantom region of the dummy
 * file and the lock is applied to file zero (whose control structure is
 * pointed to by the global variable key).
 */
	if (knum->clstyp != VAT_CLOSE)
		node /= knum->recsiz;
	node |= ((LONG) knum->filnum << LEFT_SHIFT);
	if (ctseek(ct_key,node))
		return(uerr(SEEK_ERR));
	if (locking(ct_key->fd,LK_LOCK,1L))
		return(uerr(LNOD_ERR));
	else {
		knum->lokcnt++;
		return(NO_ERROR);
	}
#endif
#else
	return(NO_ERROR);
#endif

}

/* --------------------------------------------------------------------
   UNLOCK index file node
*/

COUNT UNLOCK(node,knum)
LONG node;
PFAST KEYFILE *knum;
{

#ifdef FPUTFGET
	knum -= knum->kmem;
	if (!(knum->flmode & NONEXCLUSIVE))
		return(NO_ERROR);

	if (knum->lokcnt)
		knum->lokcnt--;
#ifndef PHANTOM
	if (ctseek(knum,node))
		return(uerr(SEEK_ERR));
	if (locking(knum->fd,LK_UNLCK,1L))
		return(uerr(UNOD_ERR));
#else
	if (knum->clstyp != VAT_CLOSE)
		node /= knum->recsiz;
	node |= ((LONG) knum->filnum << LEFT_SHIFT);
	if (ctseek(ct_key,node))
		return(uerr(SEEK_ERR));
	if (locking(ct_key->fd,LK_UNLCK,1L))
		return(uerr(UNOD_ERR));
#endif
	return(NO_ERROR);
#else
	return(NO_ERROR);
#endif

}


/* --------------------------------------------------------------------
   LOCK data record
 */

COUNT DLOCK(recbyt,dnum)		/* recbyt == 0  => header record */
POINTER		recbyt;
PFAST DATFILE  *dnum;
{

/*
 * the data record locks are NOT guaranteed to be deadlock free.  Therefore,
 * except for the header record which is only locked by internal c-tree 
 * maintenance requests, the data record lock routine returns instead of
 * sleeping when a lock is denied because of a competing lock.  The application
 * program must deal with the appropriate action if a lock is denied to a
 * data record.
 */

#ifdef FPUTFGET
#ifndef PHANTOM
	COUNT lokreq;

	if (!(dnum->flmode & NONEXCLUSIVE))
		return(NO_ERROR);

	if (ctseek(dnum,recbyt))
		return(uerr(SEEK_ERR));

	if (recbyt == DRNZERO)
		lokreq = LK_LOCK; /* allow wait for lock on header only */
	else
		lokreq = LK_NBLCK; /* return if lock not available	*/

	if (locking(dnum->fd,lokreq,1L))
		return(uerr(DLOK_ERR));
	else {
		dnum->lokcnt++;
		return(NO_ERROR);
	}
#else
	COUNT lokreq;

	if (!(dnum->flmode & NONEXCLUSIVE))
		return(NO_ERROR);

	if (recbyt == DRNZERO)
		lokreq = LK_LOCK;	/* sleep if lock is not available  */
	else
		lokreq = LK_NBLCK;	/* return if lock is not available */
	if (dnum->clstyp != VAT_CLOSE)
		recbyt /= dnum->reclen;
	recbyt |= ((LONG) dnum->filnum << LEFT_SHIFT);
	if (ctseek(ct_key,recbyt))
		return(uerr(SEEK_ERR));
	if (locking(ct_key->fd,lokreq,1L))
		return(uerr(DLOK_ERR));
	else {
		dnum->lokcnt++;
		return(NO_ERROR);
	}
#endif
#else
	return(NO_ERROR);
#endif

}

COUNT RLOCK(recbyt,dnum)
POINTER		recbyt;
PFAST DATFILE  *dnum;
{
	return(NO_ERROR);
}

/* --------------------------------------------------------------------
   UNLOCK data record
*/

COUNT UDLOCK(recbyt,dnum)
POINTER		recbyt;
PFAST DATFILE  *dnum;
{

#ifdef FPUTFGET
	if (!(dnum->flmode & NONEXCLUSIVE))
		return(NO_ERROR);

	if (dnum->lokcnt)
		dnum->lokcnt--;
#ifndef PHANTOM
	if (ctseek(dnum,recbyt))
		return(uerr(SEEK_ERR));
	if (locking(dnum->fd,LK_UNLCK,1L))
		return(uerr(UDLK_ERR));
#else
	if (dnum->clstyp != VAT_CLOSE)
		recbyt /= dnum->reclen;
	recbyt |= ((LONG) dnum->filnum << LEFT_SHIFT);
	if (ctseek(ct_key,recbyt))
		return(uerr(SEEK_ERR));
	if (locking(ct_key->fd,LK_UNLCK,1L))
		return(uerr(UDLK_ERR));
#endif
	return(NO_ERROR);

#else
	return(NO_ERROR);
#endif

}

#endif

/* end of ctclib.xnv */
